
from vnpy.event import Event, EventEngine
from vnpy.rpc import RpcServer
from vnpy.trader.engine import BaseEngine, MainEngine
from vnpy.trader.object import TickData, ContractData, LogData, SubscribeRequest
from vnpy.trader.event import EVENT_TICK


APP_NAME = "ExcelRtd"

EVENT_RTD_LOG = "eRtdLog"

REP_ADDRESS = "tcp://*:9001"
PUB_ADDRESS = "tcp://*:9002"


class RtdEngine(BaseEngine):
    """
    The engine for managing RTD objects and data update.
    """

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__(main_engine, event_engine, APP_NAME)

        self.server: RpcServer = RpcServer()
        self.server.register(self.subscribe)
        self.server.register(self.write_log)
        self.server.start(REP_ADDRESS, PUB_ADDRESS)

        self.subscribed: set[str] = set()

        self.register_event()

    def register_event(self) -> None:
        """
        Register event handler.
        """
        self.event_engine.register(EVENT_TICK, self.process_tick_event)

    def process_tick_event(self, event: Event) -> None:
        """
        Process tick event and update related RTD value.
        """
        tick: TickData = event.data
        self.server.publish("tick", tick)

    def write_log(self, msg: str) -> None:
        """
        Output RTD related log message.
        """
        log: LogData = LogData(msg=msg, gateway_name=APP_NAME)
        event: Event = Event(EVENT_RTD_LOG, log)
        self.event_engine.put(event)

    def subscribe(self, vt_symbol: str) -> None:
        """
        Subscribe tick data update.
        """
        contract: ContractData | None = self.main_engine.get_contract(vt_symbol)
        if not contract:
            return

        if vt_symbol in self.subscribed:
            return
        self.subscribed.add(vt_symbol)

        req: SubscribeRequest = SubscribeRequest(
            contract.symbol,
            contract.exchange
        )
        self.main_engine.subscribe(req, contract.gateway_name)

    def close(self) -> None:
        """"""
        self.server.stop()
        self.server.join()



from collections import defaultdict
from typing import Any

from pyxll import RTD, xl_func

from vnpy.rpc import RpcClient
from vnpy.trader.object import TickData


REQ_ADDRESS = "tcp://localhost:9001"
SUB_ADDRESS = "tcp://localhost:9002"


rtd_client: "RtdClient" | None = None


class ObjectRtd(RTD):
    """
    RTD proxy for object in Python.
    """

    def __init__(self, engine: "RtdClient", name: str, field: str) -> None:
        """Constructor"""
        super().__init__(value=0)

        self.engine: RtdClient = engine
        self.name: str = name
        self.field: str = field
        self.value: Any = 0

    def connect(self) -> None:
        """
        Callback when excel cell rtd is connected.
        """
        self.engine.add_rtd(self)

    def disconnect(self) -> None:
        """
        Callback when excel cell rtd is disconncted.
        """
        self.engine.remove_rtd(self)

    def update(self, data: object) -> None:
        """
        Update value in excel cell.
        """
        new_value = getattr(data, self.field, "N/A")

        if new_value != self.value:
            self.value = new_value


class RtdClient(RpcClient):
    """
    The engine for managing RTD objects and data update.
    """

    def __init__(self) -> None:
        """"""
        super().__init__()

        self.rtds: dict[str, set[ObjectRtd]] = defaultdict(set)

        global rtd_client
        rtd_client = self

    def callback(self, topic: str, data: object) -> None:
        """"""
        tick: TickData = data
        buf: set[ObjectRtd] = self.rtds[tick.vt_symbol]

        for rtd in buf:
            rtd.update(tick)

    def add_rtd(self, rtd: ObjectRtd) -> None:
        """
        Add a new RTD into the engine..
        """
        buf: set[ObjectRtd] = self.rtds[rtd.name]
        buf.add(rtd)
        self.write_log(f"新增RTD连接：{rtd.name} {rtd.field}")

        # Auto subscribe tick data
        self.subscribe(rtd.name)

    def remove_rtd(self, rtd: ObjectRtd) -> None:
        """
        Remove an existing RTD from the engine.
        """
        buf: set[ObjectRtd] = self.rtds[self.name]
        if self in buf:
            buf.remove(rtd)
            self.write_log(f"移除RTD连接：{rtd.name} {rtd.field}")


def init_client() -> None:
    """Initialize vnpy rtd client"""
    global rtd_client
    rtd_client = RtdClient()
    rtd_client.subscribe_topic("")
    rtd_client.start(REQ_ADDRESS, SUB_ADDRESS)


@xl_func("string vt_symbol, string field: rtd")    # type: ignore
def rtd_tick_data(vt_symbol: str, field: str) -> ObjectRtd:
    """
    Return the streaming value of the tick data field.
    """
    if not rtd_client:
        init_client()

    rtd = ObjectRtd(rtd_client, vt_symbol, field)  # type: ignore
    return rtd



from pathlib import Path

from vnpy.event import EventEngine, Event
from vnpy.trader.engine import MainEngine
from vnpy.trader.ui import QtWidgets, QtCore
from vnpy.trader.object import LogData
from ..engine import APP_NAME, EVENT_RTD_LOG, BaseEngine


class RtdManager(QtWidgets.QWidget):
    """"""
    signal_log: QtCore.Signal = QtCore.Signal(Event)

    def __init__(self, main_engine: MainEngine, event_engine: EventEngine) -> None:
        """"""
        super().__init__()

        self.main_engine: MainEngine = main_engine
        self.event_engine: EventEngine = event_engine
        self.rm_engine: BaseEngine = main_engine.get_engine(APP_NAME)

        self.init_ui()
        self.register_event()

    def init_ui(self) -> None:
        """
        Init widget ui components.
        """
        self.setWindowTitle("Excel RTD")
        self.resize(600, 600)

        module_path: Path = Path(__file__).parent.parent
        client_path: Path = module_path.joinpath("vnpy_rtd.py")
        self.client_line: QtWidgets.QLineEdit = QtWidgets.QLineEdit(str(client_path))
        self.client_line.setReadOnly(True)

        copy_button: QtWidgets.QPushButton = QtWidgets.QPushButton("复制")
        copy_button.clicked.connect(self.copy_client_path)

        self.log_monitor: QtWidgets.QTextEdit = QtWidgets.QTextEdit()
        self.log_monitor.setReadOnly(True)

        self.port_label: QtWidgets.QLabel = QtWidgets.QLabel(
            "使用Socket端口：请求回应9001、广播推送9002"
        )

        hbox: QtWidgets.QHBoxLayout = QtWidgets.QHBoxLayout()
        hbox.addWidget(self.client_line)
        hbox.addWidget(copy_button)

        vbox: QtWidgets.QVBoxLayout = QtWidgets.QVBoxLayout()
        vbox.addLayout(hbox)
        vbox.addWidget(self.log_monitor)
        vbox.addWidget(self.port_label)
        self.setLayout(vbox)

    def register_event(self) -> None:
        """
        Register event handler.
        """
        self.signal_log.connect(self.process_log_event)

        self.event_engine.register(EVENT_RTD_LOG, self.signal_log.emit)

    def process_log_event(self, event: Event) -> None:
        """
        Show log message in monitor.
        """
        log: LogData = event.data

        msg: str = f"{log.time}: {log.msg}"
        self.log_monitor.append(msg)

    def copy_client_path(self) -> None:
        """
        Copy path of client python file to clipboard.
        """
        self.client_line.selectAll()
        self.client_line.copy()
